package ru.dbotthepony.kstarbound.util

import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.RegistryObject
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.item.IItemDefinition

class ItemDescriptor private constructor(item: RegistryObject<IItemDefinition>?, count: Long, val parameters: JsonObject, marker: Unit) {
	constructor(item: RegistryObject<IItemDefinition>, count: Long = 1L, parameters: JsonObject = JsonObject()) : this(item, count, parameters, Unit)

	var item: RegistryObject<IItemDefinition>? = item
		private set

	var count = count
		set(value) {
			if (field == 0L)
				return

			field = value.coerceAtLeast(0L)

			if (field == 0L) {
				item = null
			}
		}

	val isEmpty: Boolean
		get() = count <= 0 || item == null

	val isNotEmpty: Boolean
		get() = count > 0 && item != null

	fun grow(amount: Long) {
		count += amount
	}

	fun shrink(amount: Long) {
		count -= amount
	}

	/**
	 * Возвращает null если этот предмет пуст
	 */
	fun conciseToNull(): ItemDescriptor? {
		if (isEmpty) {
			return null
		} else {
			return this
		}
	}

	fun mergeFrom(other: ItemDescriptor, simulate: Boolean) {
		if (isStackable(other)) {
			val newCount = (count + other.count).coerceAtMost(item!!.value.maxStack)
			val diff = newCount - count
			other.count -= diff

			if (!simulate)
				count = newCount
		}
	}

	fun lenientEquals(other: Any?): Boolean {
		if (other !is ItemDescriptor)
			return false

		if (isEmpty)
			return other.isEmpty

		return other.count == count && other.item == item
	}

	fun isStackable(other: ItemDescriptor): Boolean {
		return count != 0L && other.count != 0L && item!!.value.maxStack < count && other.item == item && other.parameters == parameters
	}

	override fun equals(other: Any?): Boolean {
		if (other !is ItemDescriptor)
			return false

		if (isEmpty)
			return other.isEmpty

		return other.count == count && other.item == item && other.parameters == parameters
	}

	// мы не можем делать хеш из count и parameters так как они изменяемы
	override fun hashCode(): Int {
		return item.hashCode()
	}

	override fun toString(): String {
		if (isEmpty)
			return "ItemDescriptor.EMPTY"

		return "ItemDescriptor[${item!!.value.itemName}, count = $count, params = $parameters]"
	}

	fun copy(): ItemDescriptor {
		if (isEmpty)
			return this

		return ItemDescriptor(item, count, parameters.deepCopy(), Unit)
	}

	fun toJson(): JsonObject? {
		if (isEmpty) {
			return null
		}

		return JsonObject().also {
			it.add("name", JsonPrimitive(item!!.value.itemName))
			it.add("count", JsonPrimitive(count))
			it.add("parameters", parameters.deepCopy())
		}
	}

	class Adapter(val starbound: Starbound) : TypeAdapter<ItemDescriptor>() {
		override fun write(out: JsonWriter, value: ItemDescriptor?) {
			val json = value?.toJson()

			if (json == null)
				out.nullValue()
			else
				TypeAdapters.JSON_ELEMENT.write(out, json)
		}

		override fun read(`in`: JsonReader): ItemDescriptor {
			if (`in`.peek() == JsonToken.NULL)
				return EMPTY

			return starbound.item(TypeAdapters.JSON_ELEMENT.read(`in`))
		}
	}

	companion object {
		@JvmField
		val EMPTY = ItemDescriptor(null, 0L, JsonObject(), Unit)
	}
}
